#################
#Dec 5, 2014
#Phil Boonstra
#Contains the primary simulation functions. 
#This may be called directly by the user but is primarily
#intended to be called from within a 'MisclassSims_X.R' script
################


#Simulate from a mixture of exponential random variables
#(Used in random parameter setting)
rmixexp = function(M,rates,probs) {
  clust = sample(length(rates),M,T,probs);
  x = rexp(M,rate=rates[clust]);
}

#Prospective logistic log-likelihood of a 2x2x2 table under a given parameter configuration
logLik_GxE = function(r100,r101,r110,r111,r000,r001,r010,r011,beta0,betaG,betaE,betaGE) {
  beta0*(r100+r101+r110+r111) + 
    betaG*(r110+r111) + 
    betaE*(r101+r111) + 
    betaGE*r111 - 
    log(1+exp(beta0))*(r000+r100) -
    log(1+exp(beta0+betaG))*(r010+r110) - 
    log(1+exp(beta0+betaE))*(r001+r101) -
    log(1+exp(beta0+betaE+betaG+betaGE))*(r011+r111);
}

#Solve for theta0 given PG,PE, and thetaGE
solve_for_exp_theta0 = function(exp_thetaGE,prob_G,prob_E) {
  A = -exp_thetaGE*(1-prob_G);
  B = prob_G+prob_E + exp_thetaGE*(prob_G-prob_E)-1;
  root1 = (-B-sqrt(B^2-4*A*prob_G))/(2*A);
  root2 = (-B+sqrt(B^2-4*A*prob_G))/(2*A);
  pmax(root1,root2);
}

#Used to solve for betaG given the marginal exposures and genetic ORs, 
#the GEI parameter, and control probability vector. 
#(Used in random parameter setting)
betaG_func = function(betaG,exp_alphaE,exp_alphaG,exp_betaGE,p000,p001,p010,p011) {
  exp_betaE = exp_alphaE/((p000+p010)/(p001+p011))/((exp(betaG)*exp_betaGE*p011+p001)/(exp(betaG)*p010+p000));
  betaG - log(exp_alphaG) + log((p000+p001)/(p010+p011)) + log((exp_betaE*exp_betaGE*p011+p010)/(exp_betaE*p001+p000));
} 

##Use CGEN functions to get test statistics
##
cml.test = function(ncase,ncontrol,r100,r101,r110,r111,r000,r001,r010,r011) {
  
  Dis = c(rep(0,ncontrol),rep(1,ncase)); 
  G = c(rep(0,r000+r001),rep(1,r010+r011),rep(0,r100+r101),rep(1,r110+r111))
  E = c(rep(0,r000),rep(1,r001),rep(0,r010),rep(1,r011),rep(0,r100),rep(1,r101),rep(0,r110),rep(1,r111))
  dat = data.frame(cbind(Dis,G,E));
  
  foo = snp.logistic(dat,"Dis","G","E","E")
  
  list(kraft_cml_stat = drop(foo$EB$parms[c("G","G:E")]%*%solve(foo$EB$cov[c("G","G:E"),c("G","G:E")])%*%foo$EB$parms[c("G","G:E")]),
       cnstr_cc_cml_stat = sum(foo$EB$parms[c("G","G:E")])^2/(foo$EB$cov["G","G"]+foo$EB$cov["G:E","G:E"]+2*foo$EB$cov["G","G:E"]));
}

GwaspowerRand.wrapper = function(params) {
  GwaspowerRand(niter=params$niter,
                ncase=params$ncase,ncontrol=params$ncontrol,nsnp=params$nsnp,
                gw_alpha=params$gw_alpha,
                screen_alpha_lax=params$screen_alpha_lax,screen_alpha_rigid=params$screen_alpha_rigid,
                prob_E=params$prob_E,
                marg_OR_E=params$marg_OR_E,
                sens_E_case=params$sens_E_case,spec_E_case=params$spec_E_case,
                sens_E_control=params$sens_E_control,spec_E_control=params$spec_E_control,			
                random_ORs = params$random_ORs,
                OR_GxE_assc_params=params$OR_GxE_assc_params,
                marg_OR_G_params = params$marg_OR_G_params,
                int_type_dist = params$int_type_dist,
                causal_maf = params$causal_maf, 
                prob_GxE_ind = params$prob_GxE_ind, GxE_assc_sd = params$GxE_assc_sd, causal_OR_GxE_assc = params$causal_OR_GxE_assc,
                null_OR_E = params$null_OR_E, causal_OR_E = params$causal_OR_E, 
                num_null_with_nonnull_OR_G = params$num_null_with_nonnull_OR_G, causal_OR_G = params$causal_OR_G, causal_OR_GxE = params$causal_OR_GxE,
                include_constrained=params$include_constrained,
                random_seed=params$random_seed,sim_number=params$sim_number)
}

GwaspowerRand<-function(niter,#Number of simulations
    ncase,ncontrol,nsnp,#Number of cases, controls, and SNPs
		gw_alpha = 5e-2,#Desired genome-wide type I error rate 
    screen_alpha_lax = 5e-2, screen_alpha_rigid = 5e-4,#Two choices of screening alphas are allowed (the lax is assumed to be less stringent, i.e. larger)
    prob_E = 0.5,#Exposure prevalance
		marg_OR_E = 1.2,#Marginal exposure OR
		sens_E_case = 1.0,spec_E_case = 1.0,#Sensitivity and specificty of E among cases
		sens_E_control = sens_E_case,spec_E_control = spec_E_case,#Sensitivity and specificty of E among controls
    random_ORs = T,#Generate from random parameter setting or fixed parameter setting?
    #These are the arguments that control the novel random parameter (OR) case
    #These are ignored if random_ORs == F
		OR_GxE_assc_params = list(rate1=2*16034,rate2=2*2044,p1=0.925),#G-E association in controls comes from mixture of exponentials
    marg_OR_G_params = list(rate1=16034,rate2=2044,p1=0.925),#Marginal genetic OR comes from mixture of exponentials
		int_type_dist = c(1,0),#Number of quantitative, qualitative interactions desired, should sum to less than 'nsnp'
    #These are the arguments that control the traditional fixed parameter (OR) case
		#These are ignored if random_ORs == T
		causal_maf = 0.3, 
		prob_GxE_ind = 1.0,GxE_assc_sd = log(1.5)/2, causal_OR_GxE_assc = 1.0,
		null_OR_E = 1.3, causal_OR_E = 1.0, num_null_with_nonnull_OR_G = 0, causal_OR_G = 1.0, causal_OR_GxE = 1.0,
		#Should CML (retrospective likelihood) methods be fitted (will run long if nsnp is large)
    include_constrained=F,
		#Other parameters
		random_seed,sim_number) {
	
  M = nsnp;
  if(random_ORs){
    if(min(int_type_dist<0)|max(int_type_dist)<1|sum(int_type_dist)>M) {stop("'int_type_dist' must be a vector of three positive numbers such that at least one exceeds 1 and they sum to less than 'nsnp'.")}
    if(M<10) {stop("'nsnp' too small; 'nsnp' should be at least 10 in the random parameter estting")}
  } else {
    if(M==1) {stop("set 'nsnp' > 2; there should be at least one null snp");}
  }
  
  if(!is.null(random_seed)) {set.seed(as.numeric(random_seed));}
  
  #STORE RESULTS
  ge_methods = c("CC","CO","EB","MUR_LAX","MUR_RIGID","HYB_LAX","HYB_RIGID","EDGE_SS","EDGE_WT","CT1","CT2","CT3","WAS_LAX","WAS_RIGID");
  
  if(include_constrained) {
    if(M>5) {warning("Consider setting 'include_constrained=F'; code for CONSTR_CC_CML and KRAFT_CML is not vectorized and will run long when simulating many SNPs");}
    gene_discov_methods = c("CONSTR_CC","CONSTR_CC_CML","CHISQ_CC","CHISQ_CO","CHISQ_EB","KRAFT","KRAFT_CML","G_MARG");
  } else {
    gene_discov_methods = c("CONSTR_CC","CHISQ_CC","CHISQ_CO","CHISQ_EB","KRAFT","G_MARG");
  }
  gene_discov_results = matrix(0,nrow=9,ncol=length(gene_discov_methods),dimnames=list(c("GV","TP_PROB","TP_COUNT","FP_PROB","FP_COUNT","FDR","FNR","REJECT_PROB","POS_FDR"),gene_discov_methods));
  ge_results = matrix(0,nrow=9,ncol=length(ge_methods)+length(gene_discov_methods),dimnames=list(c("GV","TP_PROB","TP_COUNT","FP_PROB","FP_COUNT","FDR","FNR","REJECT_PROB","POS_FDR"),c(ge_methods,gene_discov_methods)));
  
  marg_OR_E_supplied = !is.na(marg_OR_E);
  logic_check = matrix(0,nrow=niter,ncol=2,dimnames=list(NULL,c("marg_OR_G","marg_OR_E")))
  GE_DG_cor_null = GE_DG_cor_all = numeric(niter); 
  nsnps_passed_screening = numeric(11);
  names(nsnps_passed_screening) = c("MUR_LAX","MUR_RIGID","HYB_LAX_assc_only","HYB_LAX_marg_only","HYB_LAX_both","HYB_RIGID_assc_only","HYB_RIGID_marg_only","HYB_RIGID_both","EDGE_SS","WAS_LAX","WAS_RIGID");
  
  ge_snp_types = numeric(3);
  names(ge_snp_types) = c("null","type1","type2");
  gene_discov_snp_types = numeric(4);
  names(gene_discov_snp_types) = c("null","int","marg","both"); 
  sum_effect_size = matrix(0,nrow=1,ncol=5,dimnames=list(NULL,c("OR_GxE_assc","marg_OR_G","OR_E","OR_G","OR_GxE")));
  
  #SIGNIFICANCE FOR STANDARD BONFERRONI ADJUSTMENT
  bonferroni_sig = -qnorm(gw_alpha/(2*M));
  bonferroni_sig_chisq1 = qchisq(gw_alpha/M,df=1,lower=F);
  bonferroni_sig_chisq2 = qchisq(gw_alpha/M,df=2,lower=F);
  
  #SCREENING SIGNIFICANCE LVELS FOR MURCRAY TWO-STEP
  #HYBRID, AND EDGE (TWO DIFFERENT CHOICES OF ALPHA)
  screen_sig_lax = -qnorm(screen_alpha_lax/2);	
  screen_sig_rigid = -qnorm(screen_alpha_rigid/2);	
  screen_sig_lax_chisq2 = qchisq(screen_alpha_lax,df=2,lower=F);
  
  #For reporting progress of simulation
  report_freq = ifelse(M<2e4,100,25);
  report_units = "hours";
  i=1;
  begin = Sys.time();

	for (i in 1:niter) {
	  
	  ##################################################
	  ##################################################
	  ###WORKFLOW OF A SIMULATION:
	  ###(1) GENERATE COUNTS, THEN TEST STATISTICS FOR ALL LOCI
	  ###	  CALCULATE TYPE I ERROR, THEN POWER FOR: 
	  ###Gene-Environment Methods
	  ###(2a) CASE-CONTROL
	  ###(2b) CASE-ONLY 
	  ###(2c) EMPIRICAL BAYES (Biometrics, v64, 2008)
	  ###(2d) MURCRAY TWO-STEP, 2 VERSIONS (AJE, v169, 2009)
	  ###(2e) HYBRID, 2 VERSIONS (GenEpi, v35, 2011)
	  ###(2f) EDGE, 2 VERSIONS (GenEpi, v37, 2013)
	  ###(2g) COCKTAIL, 3 VERSIONS (GenEpi, v36, 2012)
    ### CT1 = uses threshold t=1e-3 for screening, and combination EB/CC for testing (used in paper)
	  ### CT2 = uses threshold pMar<pCor for screening, and combination EB/CC for testing
	  ### CT3 = uses threshold t=1e-3 for screening, and only CC for testing
	  ###(2h) WASON TWO-STEP (AJHG, v90, 2012)
    ###Gene Discovery Methods
	  ###(3a) CONSTRAINED CC, New Method, Constrains genetic main effect to be zero
	  ###(3b) CHISQ,CC, (AJE, v176, 2012)
	  ###(3c) CHISQ,CO, (AJE, v176, 2012)
	  ###(3d) CHISQ,EB, (AJE, v176, 2012)
	  ###(3e) KRAFT,LRT, (Human Heredity, v63, 2007)
	  ###(3f) MARGINAL G
    ###(3g) CONSTRAINED CC (CML), CONSTRAINED CC based on retrospective likelihood (EB-version)
	  ###(3h) KRAFT (CML), KRAFT,LRT based on retrospecive likelihood (EB-version)
    #############	
	  ##################################################
	  
	  
	  ###############################################################
	  #Begin (1) GENERATE COUNTS, THEN TEST STATISTICS FOR ALL LOCI#
	  ###############################################################
	  
	  if(random_ORs) {#Random parameter setting -- many SNPs are causal with varying ORs
	    qa = runif(M,0.10,0.30); #Minor allele frequency
	    prob_G = qa^2 + 2*qa*(1-qa); #Hardy-Weinberg-based probability that G=1
	    GxE_assc_effect_size = do.call("rmixexp",c(list(M=M),OR_GxE_assc_params));
	    OR_GxE_assc = exp(sqrt(GxE_assc_effect_size/prob_E/(1-prob_E))*sample(c(-1,1),M,r=T));
	  } else {#Fixed parameter setting -- first SNP is causal with fixed OR and MAF, all others are null;
	    qa = c(causal_maf,runif(M-1,0.10,0.30)); #Minor allele frequency for null loci
	    prob_G = qa^2 + 2*qa*(1-qa); #Hardy-Weinberg-based probability that G=1
	    OR_GxE_assc = c(causal_OR_GxE_assc,1 + numeric(M-1))+ .Machine$double.eps^.5 ;
	    x = c(0,rbinom(M-1,1,1-prob_GxE_ind)); #determine gxe dependence of null-loci
	    OR_GxE_assc[which(x==1)] = exp(rnorm(sum(x),0,GxE_assc_sd)) ;  #Determine value of theta_GEs
	    GxE_assc_effect_size = log(OR_GxE_assc)^2*prob_E*(1-prob_E); #Induced effect size of theta_GE at each locus
	    rm(x);
	  }
	  #Used to deterine effect size of beta_GE
	  exp_theta0 = solve_for_exp_theta0(OR_GxE_assc,prob_G,prob_E);
	  
	  #Control probabilities
	  A = (1-OR_GxE_assc);
	  B = OR_GxE_assc - A * (1-prob_G-prob_E);
	  C = -OR_GxE_assc*(1-prob_E)*(1-prob_G);
	  p000 = (-B+sqrt(B^2-4*A*C))/(2*A);
	  p001 = (1-prob_G-p000); 
	  p010 = 1-prob_E-p000; 
	  p011 = 1-(p000+p001+p010);
	  prcontrol = cbind(p000,p001,p010,p011); 
	  
	  if(random_ORs) {#Random parameter setting
	    marg_effect_size = do.call("rmixexp",c(list(M=M),marg_OR_G_params));
	    while(1) {
	      #Simulate marginal genetic ORs
	      marg_OR_G = exp(sqrt(marg_effect_size/prob_G/(1-prob_G))*sample(c(-1,1),M,r=T,prob=c(1,99)));
	      #Break if sufficient numbers of ORs are generated to create desired GEIs
	      if(sum(marg_OR_G<1)>int_type_dist[2]&sum(marg_OR_G>1)>int_type_dist[1]) {break;}
	    }
	    OR_GxE = 1 + .Machine$double.eps^.5 + numeric(M);
	    int_type1_index = int_type2_index = NULL;
	    #Interaction Type I: OR_GxE  = marg_OR_G (quantitative interaction)
	    if(int_type_dist[1]>0) {
	      int_type1_index = sample(M,int_type_dist[1],p=(marg_OR_G>1)*marg_effect_size^2)
	    }
	    #Interaction Type II: OR_GxE  = 1/marg_OR_G (qualitative interaction)
	    if(int_type_dist[2]>0) {
	      int_type2_index = sample(M,int_type_dist[2],p=(marg_OR_G<1)*marg_effect_size^2)
	    } 
	    #Interaction Type I: OR_GxE  = marg_OR_G (quantitative interaction)
	    if(length(int_type1_index)>0) {
	      OR_GxE[int_type1_index] = marg_OR_G[int_type1_index];
	    }
	    #Interaction Type II: OR_GxE  = 1/marg_OR_G (qualitative interaction)
	    if(length(int_type2_index)>0) {
	      OR_GxE[int_type2_index] = 1/marg_OR_G[int_type2_index];
	    }
	    #Null category (OR_GxE=1)
	    int_type0_index = setdiff(1:M,c(int_type1_index,int_type2_index))
	    
	    #Solve for betaG, betaE based on beta_GE, marginal ORs, and control probabilities
	    foo <- mapply(FUN=uniroot,exp_alphaE=marg_OR_E,exp_alphaG=marg_OR_G,exp_betaGE=OR_GxE,p000=p000,p001=p001,p010=p010,p011=p011,MoreArgs=list(f=betaG_func,lower=log(.2),upper=log(5),tol=.Machine$double.eps^0.5))
	    OR_G = exp(as.numeric(foo[1,]))
	    OR_E = exp(log(marg_OR_E)+log(OR_G*p010+p000)+log(p011+p001)-log(OR_G*OR_GxE*p011+p001)-log(p010+p000))
	    prod_all_ORs = OR_E*OR_G*OR_GxE;
	    
	    sum_null_OR_GxE = M - sum(int_type_dist);
	    sum_int_type1 = length(int_type1_index);
	    sum_int_type2 = length(int_type2_index);
	    sum_nonnull_OR_GxE = sum(int_type_dist);
	    loci_interaction_types = numeric(M);
	    loci_interaction_types[int_type1_index] = 1;
	    loci_interaction_types[int_type2_index] = 2;
	    int_type0_index = which(loci_interaction_types==0);
	    nonnull_OR_GxE_index = which(loci_interaction_types!=0);
	    ge_snp_types = ge_snp_types + c(sum_null_OR_GxE,sum_int_type1,sum_int_type2);
	  } else {#Fixed parameter setting
	    OR_GxE = c(causal_OR_GxE,1 + .Machine$double.eps^.5 + numeric(M-1));
	    OR_G = c(causal_OR_G,exp(runif(num_null_with_nonnull_OR_G,log(1.05),log(1.2))), 1 + .Machine$double.eps^.5 + numeric(M-num_null_with_nonnull_OR_G-1));
	    if(marg_OR_E_supplied) {#supplying marginal exposure effect ensures commensurate models
	      OR_E =  exp(log(marg_OR_E) - log(OR_G*OR_GxE*p011+p001) + log(OR_G*p010+p000) - log(p010+p000) + log(p011+p001));	      
        null_OR_E = NA;
	    } else if(!is.na(causal_OR_E)&!is.na(null_OR_E)) {
	      OR_E = c(causal_OR_E,null_OR_E + .Machine$double.eps^.5 + numeric(M-1));
	      marg_OR_E = exp(log(OR_E) + log(OR_G*OR_GxE*p011+p001) - log(OR_G*p010+p000) + log(p010+p000) - log(p011+p001));	      
	    } else {
        stop("Either 'marg_OR_E' or 'causal_OR_E' and 'null_OR_E' must be provided");
	    }
      prod_all_ORs = OR_E*OR_G*OR_GxE;
	    marg_OR_G = exp(log(OR_G) + log(OR_E*OR_GxE*p011+p010) - log(OR_E*p001+p000) + log(p001+p000) - log(p011+p010));
      marg_effect_size = prob_G*(1-prob_G)*log(marg_OR_G)^2;
	    
	    int_type1_index = 1;
	    int_type2_index = NULL;
	    sum_null_OR_GxE = M - 1;
	    sum_int_type1 = length(int_type1_index);
	    sum_int_type2 = length(int_type2_index);
	    sum_nonnull_OR_GxE = 1;
	    loci_interaction_types = c(1,rep(0,M-1));
	    int_type0_index = which(loci_interaction_types==0);
	    nonnull_OR_GxE_index = which(loci_interaction_types!=0);
	    ge_snp_types = ge_snp_types + c(sum_null_OR_GxE,sum_int_type1,sum_int_type2);
	  }
	  
	  # Case probabilities are expressed in terms of control probabilities
	  #using the identity
	  #Pr(G,E|D=1) \propto Pr(G,E|D=0)*(OR_E^E * OR_G^G * OR_GxE ^(G*E))
	  prcase = cbind(p000,p001*OR_E,p010*OR_G,p011*prod_all_ORs);
	  prcase = prcase/rowSums(prcase);
	  colnames(prcase) = c("p100","p101","p110","p111");
	  
	  #Look at empiric values of marginal ORs
    empiric_marg_OR_G = as.numeric(rowSums(prcase[,c(3,4)])*rowSums(prcontrol[,c(1,2)])/rowSums(prcase[,c(1,2)])/rowSums(prcontrol[,c(3,4)]));
	  empiric_marg_OR_E = as.numeric(rowSums(prcase[,c(2,4)])*rowSums(prcontrol[,c(1,3)])/rowSums(prcase[,c(1,3)])/rowSums(prcontrol[,c(2,4)]));	    
    
	  #Check to make sure that empiric values of marginal ORs match the nominal marginal ORs
	  logic_check[i,"marg_OR_G"] = max(abs(marg_OR_G/empiric_marg_OR_G-1));
	  logic_check[i,"marg_OR_E"] = max(abs(marg_OR_E[1]/empiric_marg_OR_E-1));
	  
	  #Summaries used by gene-discovery methods
	  marg_index = which(marg_effect_size>quantile(marg_effect_size,p=0.998));
	  nonnull_gene_discov_index = sort(union(nonnull_OR_GxE_index,marg_index));
	  null_gene_discov_index = setdiff(1:M,nonnull_gene_discov_index);
	  sum_nonnull_gene_discov = length(nonnull_gene_discov_index);
	  gene_discov_snp_types = gene_discov_snp_types + c(M-sum_nonnull_gene_discov,length(setdiff(nonnull_OR_GxE_index,marg_index)),length(setdiff(marg_index,nonnull_OR_GxE_index)),length(intersect(nonnull_OR_GxE_index,marg_index)));
	  
	  OR_GxE_effect_size = prob_E*(1+1/exp_theta0/OR_GxE_assc)^(-1);
	  OR_GxE_effect_size = log(OR_GxE)^2*OR_GxE_effect_size*(1-OR_GxE_effect_size);
	  sum_effect_size = sum_effect_size + colSums(cbind(GxE_assc_effect_size,marg_effect_size,log(OR_E)^2*prob_E*(1-prob_E),log(OR_G)^2*prob_G*(1-prob_G),OR_GxE_effect_size));
	  
	  #Sample data
    counter = 0
	  while(1) {
	    freqcontrol = apply(prcontrol,1,rmultinom,n=1,size=ncontrol);
	    freqcase = apply(prcase,1,rmultinom,n=1,size=ncase);
	    if(min(c(min(freqcontrol),min(freqcase)))>=1) {break;}
	    counter = counter + 1;
      if(counter>10) {stop("Choice of ORs are yielding sparse-celled D-G-E configurations");}
	  }
	  
	  r000=as.double(freqcontrol[1,]);#D=0;G=0;E=0;
	  r001=as.double(freqcontrol[2,]);#D=0;G=0;E=1;
	  r010=as.double(freqcontrol[3,]);#D=0;G=1;E=0;
	  r011=as.double(freqcontrol[4,]);#D=0;G=1;E=1;
	  r100=as.double(freqcase[1,]);#D=1;G=0;E=1;
	  r101=as.double(freqcase[2,]);#D=1;G=0;E=1;
	  r110=as.double(freqcase[3,]);#D=1;G=1;E=0; 
	  r111=as.double(freqcase[4,]);#D=1;G=1;E=1;
	  
	  # Misclassification
	  counter = 0
	  while(1) {
	    r101_to_r100 = rbinom(M,r101,1-sens_E_case);
	    r100_to_r101 = rbinom(M,r100,1-spec_E_case);
	    
	    r100.mc = r100 - r100_to_r101 + r101_to_r100;
	    r101.mc = r101 - r101_to_r100 + r100_to_r101;
	    
	    if(min(min(r100.mc),min(r101.mc))>=1) {break;}
	    counter = counter + 1;
	    if(counter>10) {stop("Choice of `sens_E_case' and `spec_E_case' are yielding sparse-celled D-G-E configurations");}
	  }
	  
	  counter = 0
	  while(1) {	
	    r111_to_r110=rbinom(M,r111,1-sens_E_case);
	    r110_to_r111=rbinom(M,r110,1-spec_E_case);
	    
	    r110.mc = r110 - r110_to_r111 + r111_to_r110;
	    r111.mc = r111 - r111_to_r110 + r110_to_r111;
	    
	    if(min(min(r110.mc),min(r111.mc))>=1) {break;}
	    counter = counter + 1;
	    if(counter>10) {stop("Choice of `sens_E_case' and `spec_E_case' are yielding sparse-celled D-G-E configurations");}
	  }		
	  
	  counter = 0
	  while(1) {	
	    r001_to_r000=rbinom(M,r001,1-sens_E_control);
	    r000_to_r001=rbinom(M,r000,1-spec_E_control);
	    
	    r000.mc = r000 - r000_to_r001 + r001_to_r000;
	    r001.mc = r001 - r001_to_r000 + r000_to_r001;
	    
	    if(min(min(r000.mc),min(r001.mc))>=1) {break;}
	    counter = counter + 1;
	    if(counter>10) {stop("Choice of `sens_E_control' and `spec_E_control' are yielding sparse-celled D-G-E configurations");}
	  }				
	  
	  counter = 0
	  while(1) {	
	    r011_to_r010=rbinom(M,r011,1-sens_E_control);
	    r010_to_r011=rbinom(M,r010,1-spec_E_control);			
	    
	    r010.mc = r010 - r010_to_r011 + r011_to_r010;
	    r011.mc = r011 - r011_to_r010 + r010_to_r011;		
	    
	    if(min(min(r010.mc),min(r011.mc))>=1) {break;}
	    counter = counter + 1;
	    if(counter>10) {stop("Choice of `sens_E_control' and `spec_E_control' are yielding sparse-celled D-G-E configurations");}
	  }	
	  
	  r100=r100.mc;
	  r101=r101.mc;
	  r110=r110.mc;
	  r111=r111.mc;
	  r000=r000.mc;
	  r001=r001.mc;
	  r010=r010.mc;
	  r011=r011.mc;	
	  
	  #CASE-ONLY
	  logor_co=log((r100*r111)/(r101*r110)); 
	  var_logor_co=(1/r100+1/r101+1/r110+1/r111);
	  zstat_co=abs(logor_co/sqrt(var_logor_co));
	  
	  #CASE-CONTROL
	  logor_cc=log((r001*r010*r100*r111)/(r000*r011*r101*r110)); 
	  var_logor_cc=(1/r000+1/r001+1/r010+1/r011)+(1/r100+1/r101+1/r110+1/r111);
	  zstat_cc=abs(logor_cc/sqrt(var_logor_cc));
	  
	  #EMPIRICAL BAYES
	  tau=log((r000*r011)/(r001*r010));
	  var_tau=(1/r000+1/r001+1/r010+1/r011);
	  vec_prop=((var_logor_cc)/((tau)^2+var_logor_cc))*(logor_co)+ (((tau)^2)/((tau)^2+var_logor_cc))*(logor_cc);
	  k1=((tau)^2)*(3*var_logor_cc+(tau)^2)/((var_logor_cc+(tau)^2)^2);
	  varprop=var_logor_co+((k1)^2)*var_tau;
	  zstat_eb=abs(vec_prop/sqrt(varprop));
	  
	  #MURCRAY TWO-STEP (THIS IS ALSO USED IN SCREENING STEP OF HYBRID, EDGE & COCKTAIL)
	  logor_assc = log(((r000+r100)*(r011+r111))/((r010+r110)*(r001+r101)));
	  sd_logor_assc = sqrt(1/(r000+r100)+1/(r011+r111)+1/(r010+r110) +1/(r001+r101));
	  screen_assc = abs(logor_assc/sd_logor_assc);# screen for G-E association
	  
	  #HYBRID (ALSO USED IN SCREENING STEP OF EDGE & COCKTAIL)
	  logor_marg = log(((r000+r001)*(r110+r111))/((r100+r101)*(r010+r011)));
	  sd_logor_marg = sqrt(1/(r000+r001)+1/(r110+r111)+1/(r100+r101) +1/(r010+r011));
	  screen_marg = abs(logor_marg/sd_logor_marg);#screen for marginal D-G association
	  
	  #EDGE
	  screen_marg_sq = screen_marg^2;
	  screen_assc_sq = screen_assc^2;
	  GE_DG_cor_null[i] = cor(screen_assc_sq[int_type0_index],screen_marg_sq[int_type0_index]);
	  GE_DG_cor_all[i] = cor(screen_assc_sq,screen_marg_sq);
	  screen_joint_sq = screen_marg_sq + screen_assc_sq;
    
    #WASON TS
    logor_wason = logor_co + (var_logor_co/var_tau)*tau;
    sd_logor_wason = sqrt(var_logor_co*(1+var_logor_co/var_tau));
    screen_wason = abs(logor_wason/sd_logor_wason);
	  
	  #MARGINAL + GE
	  chisq_cc_stat = screen_marg_sq + zstat_cc^2
	  chisq_co_stat = screen_marg_sq + zstat_co^2
	  chisq_eb_stat = screen_marg_sq + zstat_eb^2
	  
	  #KRAFT LRT (JOINT TEST FOR MAIN AND INTERACTION)
	  beta0_full = log(r100/r000);
	  betaG_full = log((r000*r110)/(r100*r010)); 
	  betaE_full = log((r000*r101)/(r100*r001));
	  betaGE_full = logor_cc;
	  beta0_reduced = log((r100+r110)/(r000+r010));
	  betaE_reduced = log((r000+r010)*(r101+r111)/(r100+r110)/(r011+r001));
	  kraft_lrt_stat =  2*(logLik_GxE(r100,r101,r110,r111,r000,r001,r010,r011,beta0_full,betaG_full,betaE_full,betaGE_full)-
	                         logLik_GxE(r100,r101,r110,r111,r000,r001,r010,r011,beta0_reduced,0,betaE_reduced,0))
	  
	  #CONSTRAINED CC TEST
	  gamma0_full = beta0_reduced;
	  gammaE_full = log(r101*(r000+r010)/r001/(r100+r110));
	  gammaGE_full = log(r111*r001/r011/r101);
	  constr_cc_stat = 2*(logLik_GxE(r100,r101,r110,r111,r000,r001,r010,r011,gamma0_full,0,gammaE_full,gammaGE_full)-
	                        logLik_GxE(r100,r101,r110,r111,r000,r001,r010,r011,beta0_reduced,0,betaE_reduced,0))
	  
    if(include_constrained) {#Will run long if M is large
      ##Optional additional tests:
      ##Retrospective joint test (kraft_cml_test);
      ##Retrospective contrained test (cnstr_cc_cml_test);
      kraft_cml_stat = cnstr_cc_cml_stat = numeric(M);
      for(m in 1:M) {
        foo = cml.test(ncase,ncontrol,r100[m],r101[m],r110[m],r111[m],r000[m],r001[m],r010[m],r011[m]);
        kraft_cml_stat[m] = foo$kraft_cml_stat;
        cnstr_cc_cml_stat[m] = foo$cnstr_cc_cml_stat;
      }
    }
    
	  #Clean up
	  rm(freqcontrol,freqcase,
	     r100,r100.mc,
	     r101,r101.mc,
	     r110,r110.mc,
	     r111,r111.mc,
	     r000,r000.mc,
	     r001,r001.mc,
	     r010,r010.mc,
	     r011,r011.mc,	
	     logor_co,var_logor_co,
	     logor_cc,var_logor_cc,
	     tau,var_tau,vec_prop,k1,varprop,
	     logor_assc,sd_logor_assc,
       logor_wason,sd_logor_wason,
	     logor_marg,sd_logor_marg,screen_marg_sq,
	     beta0_full,betaG_full,betaE_full,betaGE_full,beta0_reduced,betaE_reduced,
	     gamma0_full,gammaE_full,gammaGE_full);
	  
	  #################################################################
	  #END (1) GENERATE COUNTS, THEN TEST STATISTICS 		#
	  #BEGIN (2a) CASE-CONTROL ANALYSIS				#
	  #################################################################
	  all_found = zstat_cc>bonferroni_sig;
	  sum_all_found = sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  ge_results["GV","CC"] = ge_results["GV","CC"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","CC"] = ge_results["FP_COUNT","CC"] + sum_null_found; 
	  ge_results["TP_COUNT","CC"] = ge_results["TP_COUNT","CC"] + sum_nonnull_found;
	  ge_results["FDR","CC"] = ge_results["FDR","CC"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","CC"] = ge_results["FNR","CC"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","CC"] = ge_results["FP_PROB","CC"] + (sum_null_found>0); 
	  ge_results["TP_PROB","CC"] = ge_results["TP_PROB","CC"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","CC"] = ge_results["REJECT_PROB","CC"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (2a) CASE-CONTROL ANALYSIS					#
	  #BEGIN (2b) CASE-ONLY ANALYSIS					#
	  #################################################################
	  all_found = zstat_co>bonferroni_sig;
	  sum_all_found = sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  ge_results["GV","CO"] = ge_results["GV","CO"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","CO"] = ge_results["FP_COUNT","CO"] + sum_null_found; 
	  ge_results["TP_COUNT","CO"] = ge_results["TP_COUNT","CO"] + sum_nonnull_found;
	  ge_results["FDR","CO"] = ge_results["FDR","CO"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","CO"] = ge_results["FNR","CO"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","CO"] = ge_results["FP_PROB","CO"] + (sum_null_found>0); 
	  ge_results["TP_PROB","CO"] = ge_results["TP_PROB","CO"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","CO"] = ge_results["REJECT_PROB","CO"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (2b) CASE-ONLY ANALYSIS					#
	  #BEGIN (2c) EMPIRICAL BAYES ANALYSIS 				#
	  #################################################################
	  all_found = zstat_eb>bonferroni_sig;
	  sum_all_found = sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  ge_results["GV","EB"] = ge_results["GV","EB"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","EB"] = ge_results["FP_COUNT","EB"] + sum_null_found; 
	  ge_results["TP_COUNT","EB"] = ge_results["TP_COUNT","EB"] + sum_nonnull_found;
	  ge_results["FDR","EB"] = ge_results["FDR","EB"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","EB"] = ge_results["FNR","EB"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","EB"] = ge_results["FP_PROB","EB"] + (sum_null_found>0); 
	  ge_results["TP_PROB","EB"] = ge_results["TP_PROB","EB"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","EB"] = ge_results["REJECT_PROB","EB"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (2c) EMPIRICAL BAYES ANALYSIS				#
	  #BEGIN (2d) MURCRAY TWO-STEP					#
	  #################################################################
	  
	  #Murcray (lax screening);
	  passed_assc_screen = screen_assc > screen_sig_lax;
	  sum_passed_assc_screen = sum(passed_assc_screen);
	  
	  mur_inter_sig = ifelse(sum_passed_assc_screen>0,-qnorm(gw_alpha/sum_passed_assc_screen/2),Inf)#determine testing alpha
	  all_found = (zstat_cc > mur_inter_sig) & passed_assc_screen;
	  sum_all_found =  sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  nsnps_passed_screening["MUR_LAX"] = nsnps_passed_screening["MUR_LAX"] + sum_passed_assc_screen;
	  ge_results["GV","MUR_LAX"] = ge_results["GV","MUR_LAX"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","MUR_LAX"] = ge_results["FP_COUNT","MUR_LAX"] + sum_null_found; 
	  ge_results["TP_COUNT","MUR_LAX"] = ge_results["TP_COUNT","MUR_LAX"] + sum_nonnull_found;
	  ge_results["FDR","MUR_LAX"] = ge_results["FDR","MUR_LAX"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","MUR_LAX"] = ge_results["FNR","MUR_LAX"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","MUR_LAX"] = ge_results["FP_PROB","MUR_LAX"] + (sum_null_found>0); 
	  ge_results["TP_PROB","MUR_LAX"] = ge_results["TP_PROB","MUR_LAX"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","MUR_LAX"] = ge_results["REJECT_PROB","MUR_LAX"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(passed_assc_screen,sum_passed_assc_screen,mur_inter_sig,all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #Murcray (rigid screening);
	  passed_assc_screen = screen_assc > screen_sig_rigid;
	  sum_passed_assc_screen = sum(passed_assc_screen);
	  
	  mur_inter_sig = ifelse(sum_passed_assc_screen>0,-qnorm(gw_alpha/sum_passed_assc_screen/2),Inf)#determine second-step assc alpha, equal parts to each
	  all_found = (zstat_cc > mur_inter_sig) & passed_assc_screen;
	  sum_all_found =  sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  nsnps_passed_screening["MUR_RIGID"] = nsnps_passed_screening["MUR_RIGID"] + sum_passed_assc_screen;
	  ge_results["GV","MUR_RIGID"] = ge_results["GV","MUR_RIGID"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","MUR_RIGID"] = ge_results["FP_COUNT","MUR_RIGID"] + sum_null_found; 
	  ge_results["TP_COUNT","MUR_RIGID"] = ge_results["TP_COUNT","MUR_RIGID"] + sum_nonnull_found;
	  ge_results["FDR","MUR_RIGID"] = ge_results["FDR","MUR_RIGID"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","MUR_RIGID"] = ge_results["FNR","MUR_RIGID"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","MUR_RIGID"] = ge_results["FP_PROB","MUR_RIGID"] + (sum_null_found>0); 
	  ge_results["TP_PROB","MUR_RIGID"] = ge_results["TP_PROB","MUR_RIGID"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","MUR_RIGID"] = ge_results["REJECT_PROB","MUR_RIGID"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(passed_assc_screen,sum_passed_assc_screen,mur_inter_sig,all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (2d) MURCRAY TWO-STEP					#
	  #BEGIN (2e) HYBRID						#
	  #################################################################
	  
	  #Hybrid  (lax screening);
	  passed_both_screens = screen_assc > screen_sig_lax & screen_marg > screen_sig_lax#screen for both assc. and marg. 
	  sum_passed_both_screens = sum(passed_both_screens);
	  passed_assc_screen = screen_assc > screen_sig_lax#assc.
	  sum_passed_assc_screen = sum(passed_assc_screen);
	  passed_marg_screen = screen_marg > screen_sig_lax#marg.
	  sum_passed_marg_screen = sum(passed_marg_screen);
	  
	  hyb_inter_sig_assc = ifelse(sum_passed_assc_screen>0,-qnorm(.5*gw_alpha/sum_passed_assc_screen/2),Inf)#determine second-step assc alpha, equal parts to each
	  hyb_inter_sig_marg = ifelse(sum_passed_marg_screen>0,-qnorm(.5*gw_alpha/sum_passed_marg_screen/2),Inf)#determine second-step marg alpha
	  
	  all_found = 
	    (zstat_cc > hyb_inter_sig_assc & passed_assc_screen & !passed_both_screens)|
	    (zstat_cc > hyb_inter_sig_marg & passed_marg_screen & !passed_both_screens)|
	    (zstat_cc > min(hyb_inter_sig_assc,hyb_inter_sig_marg) & passed_both_screens)
	  sum_all_found = sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  nsnps_passed_screening[c("HYB_LAX_assc_only","HYB_LAX_marg_only","HYB_LAX_both")] = nsnps_passed_screening[c("HYB_LAX_assc_only","HYB_LAX_marg_only","HYB_LAX_both")] + c(sum_passed_assc_screen-sum_passed_both_screens,sum_passed_marg_screen-sum_passed_both_screens,sum_passed_both_screens);
	  
	  #Summaries of only interaction SNPs found
	  ge_results["GV","HYB_LAX"] = ge_results["GV","HYB_LAX"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","HYB_LAX"] = ge_results["FP_COUNT","HYB_LAX"] + sum_null_found; 
	  ge_results["TP_COUNT","HYB_LAX"] = ge_results["TP_COUNT","HYB_LAX"] + sum_nonnull_found;
	  ge_results["FDR","HYB_LAX"] = ge_results["FDR","HYB_LAX"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","HYB_LAX"] = ge_results["FNR","HYB_LAX"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","HYB_LAX"] = ge_results["FP_PROB","HYB_LAX"] + (sum_null_found>0); 
	  ge_results["TP_PROB","HYB_LAX"] = ge_results["TP_PROB","HYB_LAX"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","HYB_LAX"] = ge_results["REJECT_PROB","HYB_LAX"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(passed_assc_screen,sum_passed_assc_screen,passed_marg_screen,sum_passed_marg_screen,passed_both_screens,sum_passed_both_screens,all_found,hyb_inter_sig_assc,hyb_inter_sig_marg,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #Hybrid  (rigid screening);
	  passed_both_screens = screen_assc > screen_sig_rigid & screen_marg > screen_sig_rigid#screen for both assc. and marg. 
	  sum_passed_both_screens = sum(passed_both_screens);
	  passed_assc_screen = screen_assc > screen_sig_rigid#assc.
	  sum_passed_assc_screen = sum(passed_assc_screen);
	  passed_marg_screen = screen_marg > screen_sig_rigid#marg.
	  sum_passed_marg_screen = sum(passed_marg_screen);
	  
	  hyb_inter_sig_assc = ifelse(any(passed_assc_screen),-qnorm(.5*gw_alpha/sum(passed_assc_screen)/2),Inf)#determine second-step assc alpha, equal parts to each
	  hyb_inter_sig_marg = ifelse(any(passed_marg_screen),-qnorm(.5*gw_alpha/sum(passed_marg_screen)/2),Inf)#determine second-step marg alpha
	  
	  all_found = 
	    (zstat_cc > hyb_inter_sig_assc & passed_assc_screen & !passed_both_screens)|
	    (zstat_cc > hyb_inter_sig_marg & passed_marg_screen & !passed_both_screens)|
	    (zstat_cc > min(hyb_inter_sig_assc,hyb_inter_sig_marg) & passed_both_screens)
	  sum_all_found = sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  nsnps_passed_screening[c("HYB_RIGID_assc_only","HYB_RIGID_marg_only","HYB_RIGID_both")] = nsnps_passed_screening[c("HYB_RIGID_assc_only","HYB_RIGID_marg_only","HYB_RIGID_both")] + c(sum_passed_assc_screen-sum_passed_both_screens,sum_passed_marg_screen-sum_passed_both_screens,sum_passed_both_screens);
	  
	  #Summaries of only interaction SNPs found
	  ge_results["GV","HYB_RIGID"] = ge_results["GV","HYB_RIGID"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","HYB_RIGID"] = ge_results["FP_COUNT","HYB_RIGID"] + sum_null_found; 
	  ge_results["TP_COUNT","HYB_RIGID"] = ge_results["TP_COUNT","HYB_RIGID"] + sum_nonnull_found;
	  ge_results["FDR","HYB_RIGID"] = ge_results["FDR","HYB_RIGID"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","HYB_RIGID"] = ge_results["FNR","HYB_RIGID"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","HYB_RIGID"] = ge_results["FP_PROB","HYB_RIGID"] + (sum_null_found>0); 
	  ge_results["TP_PROB","HYB_RIGID"] = ge_results["TP_PROB","HYB_RIGID"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","HYB_RIGID"] = ge_results["REJECT_PROB","HYB_RIGID"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(passed_assc_screen,sum_passed_assc_screen,passed_marg_screen,sum_passed_marg_screen,passed_both_screens,sum_passed_both_screens,all_found,hyb_inter_sig_assc,hyb_inter_sig_marg,sum_all_found,sum_null_found,sum_nonnull_found);
	  #################################################################
	  #END (2e) HYBRID  					#
	  #BEGIN (2f) EDGE
	  #################################################################
	  
	  #EDGE (subset approach);
	  passed_joint_screen = screen_joint_sq > screen_sig_lax_chisq2;
	  sum_passed_joint_screen = sum(passed_joint_screen);
	  
	  edge_inter_sig = ifelse(sum_passed_joint_screen>0,-qnorm(gw_alpha/sum_passed_joint_screen/2),Inf)#determine testing alpha
	  
	  all_found = (zstat_cc > edge_inter_sig) & passed_joint_screen;
	  sum_all_found =  sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  nsnps_passed_screening["EDGE_SS"] = nsnps_passed_screening["EDGE_SS"] + sum_passed_joint_screen;
	  ge_results["GV","EDGE_SS"] = ge_results["GV","EDGE_SS"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","EDGE_SS"] = ge_results["FP_COUNT","EDGE_SS"] + sum_null_found; 
	  ge_results["TP_COUNT","EDGE_SS"] = ge_results["TP_COUNT","EDGE_SS"] + sum_nonnull_found;
	  ge_results["FDR","EDGE_SS"] = ge_results["FDR","EDGE_SS"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","EDGE_SS"] = ge_results["FNR","EDGE_SS"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","EDGE_SS"] = ge_results["FP_PROB","EDGE_SS"] + (sum_null_found>0); 
	  ge_results["TP_PROB","EDGE_SS"] = ge_results["TP_PROB","EDGE_SS"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","EDGE_SS"] = ge_results["REJECT_PROB","EDGE_SS"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(passed_joint_screen,sum_passed_joint_screen,edge_inter_sig,all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  
	  #EDGE (weighted pvalue appraoch);
	  psc = pchisq(screen_joint_sq,df=2,lower=F);
	  
	  #Rank screening pvalues
	  #By construction, both categories will have
	  #same marginal numbers in each rank, only individual
	  #rankings will vary
	  grp = ceiling(log(rank(psc)/5+1)/log(2))
	  wt_pv =  numeric(M);
	  
	  for(k in 1:max(grp)) {
	    grp_index = which(grp==k);
	    wt_pv[grp_index] =
	      (gw_alpha)*(2^(-k))/length(grp_index)
	  }
	  
	  p_vals = 2*pnorm(zstat_cc,lower=F)
	  all_found = (p_vals < wt_pv);
	  
	  sum_all_found = sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  #Summaries of all causal SNPs found
	  ge_results["GV","EDGE_WT"] = ge_results["GV","EDGE_WT"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","EDGE_WT"] = ge_results["FP_COUNT","EDGE_WT"] + sum_null_found; 
	  ge_results["TP_COUNT","EDGE_WT"] = ge_results["TP_COUNT","EDGE_WT"] + sum_nonnull_found;
	  ge_results["FDR","EDGE_WT"] = ge_results["FDR","EDGE_WT"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","EDGE_WT"] = ge_results["FNR","EDGE_WT"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","EDGE_WT"] = ge_results["FP_PROB","EDGE_WT"] + (sum_null_found>0); 
	  ge_results["TP_PROB","EDGE_WT"] = ge_results["TP_PROB","EDGE_WT"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","EDGE_WT"] = ge_results["REJECT_PROB","EDGE_WT"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(psc,grp,wt_pv,k,grp_index,p_vals,all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (2f) EDGE
	  #BEGIN (2g) COCKTAIL						#
	  #################################################################
	  
	  # correlation effect/marginal pvalue
	  psc1 = psc2 = 2*pnorm(screen_assc,lower=F);
	  pMar = 2*pnorm(screen_marg,lower=F);
	  pMar_lt_c = which(pMar <= 1e-3);
	  pMar_lt_pCor = which(pMar < psc1);
	  
	  # screening pvalues
	  psc1[pMar_lt_c] = pMar[pMar_lt_c];
	  psc2[pMar_lt_pCor] = pMar[pMar_lt_pCor];
	  
	  # G*E pvalues	
	  pge1 = pge2 = pge3 = 2*pnorm(zstat_cc,lower=F);
	  peb = 2*pnorm(zstat_eb,lower=F);
	  pge1[pMar_lt_c] = peb[pMar_lt_c];
	  pge2[pMar_lt_pCor] = peb[pMar_lt_pCor];
	  
	  #Rank screening pvalues
	  #By construction, both groups will have
	  #same marginal numbers in each rank, only individual
	  #rankings will vary
	  grp1 = ceiling(log(rank(psc1)/5+1,base=2));
	  grp2 = ceiling(log(rank(psc2)/5+1,base=2));
	  wt_pv1 = wt_pv2 = numeric(M);
	  
	  for(k in 1:max(grp1))
	  {
	    grp1_index = which(grp1==k);
	    grp2_index = which(grp2==k);
	    wt_pv1[grp1_index] =
	      wt_pv2[grp2_index] = 
	      (gw_alpha)*(2^(-k))/length(grp1_index)
	  }
	  wt_pv3 = wt_pv1;#CT1 and CT3 do identical screening;
	  rm(pMar,psc1,psc2,pMar_lt_c,pMar_lt_pCor,peb,grp1,grp2,k);
	    
	  #CT 1
	  all_found = pge1 < wt_pv1;
	  sum_all_found = sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  #Summaries of all causal SNPs found
	  ge_results["GV","CT1"] = ge_results["GV","CT1"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","CT1"] = ge_results["FP_COUNT","CT1"] + sum_null_found; 
	  ge_results["TP_COUNT","CT1"] = ge_results["TP_COUNT","CT1"] + sum_nonnull_found;
	  ge_results["FDR","CT1"] = ge_results["FDR","CT1"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","CT1"] = ge_results["FNR","CT1"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","CT1"] = ge_results["FP_PROB","CT1"] + (sum_null_found>0); 
	  ge_results["TP_PROB","CT1"] = ge_results["TP_PROB","CT1"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","CT1"] = ge_results["REJECT_PROB","CT1"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  rm(pge1,wt_pv1);
	  
	  ##Cocktail 2
	  all_found = pge2 < wt_pv2; 
	  sum_all_found = sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  #Summaries of all causal SNPs found
	  ge_results["GV","CT2"] = ge_results["GV","CT2"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","CT2"] = ge_results["FP_COUNT","CT2"] + sum_null_found; 
	  ge_results["TP_COUNT","CT2"] = ge_results["TP_COUNT","CT2"] + sum_nonnull_found;
	  ge_results["FDR","CT2"] = ge_results["FDR","CT2"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","CT2"] = ge_results["FNR","CT2"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","CT2"] = ge_results["FP_PROB","CT2"] + (sum_null_found>0); 
	  ge_results["TP_PROB","CT2"] = ge_results["TP_PROB","CT2"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","CT2"] = ge_results["REJECT_PROB","CT2"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  rm(pge2,wt_pv2);
    
	  ##Cocktail 3
	  all_found = pge3 < wt_pv3; 
	  sum_all_found = sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  #Summaries of all causal SNPs found
	  ge_results["GV","CT3"] = ge_results["GV","CT3"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","CT3"] = ge_results["FP_COUNT","CT3"] + sum_null_found; 
	  ge_results["TP_COUNT","CT3"] = ge_results["TP_COUNT","CT3"] + sum_nonnull_found;
	  ge_results["FDR","CT3"] = ge_results["FDR","CT3"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","CT3"] = ge_results["FNR","CT3"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","CT3"] = ge_results["FP_PROB","CT3"] + (sum_null_found>0); 
	  ge_results["TP_PROB","CT3"] = ge_results["TP_PROB","CT3"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","CT3"] = ge_results["REJECT_PROB","CT3"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  rm(pge3,wt_pv3)
	  
	  #################################################################
	  #END (2g) COCKTAIL						#
	  #BEGIN (2h) WASON TWO-STEP  				#
	  #################################################################
	  
	  #Wason (lax screening);
	  passed_wason_screen = screen_wason > screen_sig_lax;
	  sum_passed_wason_screen = sum(passed_wason_screen);
	  
	  was_inter_sig = ifelse(sum_passed_wason_screen>0,-qnorm(gw_alpha/sum_passed_wason_screen/2),Inf)#determine testing alpha
	  all_found = (zstat_cc > was_inter_sig) & passed_wason_screen;
	  sum_all_found =  sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  nsnps_passed_screening["WAS_LAX"] = nsnps_passed_screening["WAS_LAX"] + sum_passed_wason_screen;
	  ge_results["GV","WAS_LAX"] = ge_results["GV","WAS_LAX"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","WAS_LAX"] = ge_results["FP_COUNT","WAS_LAX"] + sum_null_found; 
	  ge_results["TP_COUNT","WAS_LAX"] = ge_results["TP_COUNT","WAS_LAX"] + sum_nonnull_found;
	  ge_results["FDR","WAS_LAX"] = ge_results["FDR","WAS_LAX"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","WAS_LAX"] = ge_results["FNR","WAS_LAX"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","WAS_LAX"] = ge_results["FP_PROB","WAS_LAX"] + (sum_null_found>0); 
	  ge_results["TP_PROB","WAS_LAX"] = ge_results["TP_PROB","WAS_LAX"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","WAS_LAX"] = ge_results["REJECT_PROB","WAS_LAX"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(passed_wason_screen,sum_passed_wason_screen,was_inter_sig,all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #Murcray (rigid screening);
	  passed_wason_screen = screen_wason > screen_sig_rigid;
	  sum_passed_wason_screen = sum(passed_wason_screen);
	  
	  was_inter_sig = ifelse(sum_passed_wason_screen>0,-qnorm(gw_alpha/sum_passed_wason_screen/2),Inf)#determine second-step wason alpha, equal parts to each
	  all_found = (zstat_cc > was_inter_sig) & passed_wason_screen;
	  sum_all_found =  sum(all_found);
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  
	  nsnps_passed_screening["WAS_RIGID"] = nsnps_passed_screening["WAS_RIGID"] + sum_passed_wason_screen;
	  ge_results["GV","WAS_RIGID"] = ge_results["GV","WAS_RIGID"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","WAS_RIGID"] = ge_results["FP_COUNT","WAS_RIGID"] + sum_null_found; 
	  ge_results["TP_COUNT","WAS_RIGID"] = ge_results["TP_COUNT","WAS_RIGID"] + sum_nonnull_found;
	  ge_results["FDR","WAS_RIGID"] = ge_results["FDR","WAS_RIGID"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","WAS_RIGID"] = ge_results["FNR","WAS_RIGID"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","WAS_RIGID"] = ge_results["FP_PROB","WAS_RIGID"] + (sum_null_found>0); 
	  ge_results["TP_PROB","WAS_RIGID"] = ge_results["TP_PROB","WAS_RIGID"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","WAS_RIGID"] = ge_results["REJECT_PROB","WAS_RIGID"] + (sum_all_found>0);
	  
	  #Clean up
	  rm(passed_wason_screen,sum_passed_wason_screen,was_inter_sig,all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (2h) WASON TWO-STEP    			#
	  #################################################################
	  #Beginning of Gene-discovery methods;
	  
	  #################################################################   
	  #BEGIN (3a) CONSTRAINED CC  			#
	  #################################################################
	  all_found = constr_cc_stat>bonferroni_sig_chisq1
	  sum_all_found = sum(all_found);
	  
	  #Summaries of all GxE SNPs found
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  ge_results["GV","CONSTR_CC"] = ge_results["GV","CONSTR_CC"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","CONSTR_CC"] = ge_results["FP_COUNT","CONSTR_CC"] + sum_null_found; 
	  ge_results["TP_COUNT","CONSTR_CC"] = ge_results["TP_COUNT","CONSTR_CC"] + sum_nonnull_found;
	  ge_results["FDR","CONSTR_CC"] = ge_results["FDR","CONSTR_CC"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","CONSTR_CC"] = ge_results["FNR","CONSTR_CC"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","CONSTR_CC"] = ge_results["FP_PROB","CONSTR_CC"] + (sum_null_found>0); 
	  ge_results["TP_PROB","CONSTR_CC"] = ge_results["TP_PROB","CONSTR_CC"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","CONSTR_CC"] = ge_results["REJECT_PROB","CONSTR_CC"] + (sum_all_found>0);
	  
	  sum_null_found = sum(all_found[null_gene_discov_index]);
	  sum_nonnull_found = sum(all_found[nonnull_gene_discov_index]);
	  
	  #Summaries of gene discoveries
	  gene_discov_results["GV","CONSTR_CC"] = gene_discov_results["GV","CONSTR_CC"] + sum(all_found*marg_effect_size);
	  gene_discov_results["FP_COUNT","CONSTR_CC"] = gene_discov_results["FP_COUNT","CONSTR_CC"] + sum_null_found; 
	  gene_discov_results["TP_COUNT","CONSTR_CC"] = gene_discov_results["TP_COUNT","CONSTR_CC"] + sum_nonnull_found;
	  gene_discov_results["FDR","CONSTR_CC"] = gene_discov_results["FDR","CONSTR_CC"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  gene_discov_results["FNR","CONSTR_CC"] = gene_discov_results["FNR","CONSTR_CC"] + max(0,(sum_nonnull_gene_discov-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  gene_discov_results["FP_PROB","CONSTR_CC"] = gene_discov_results["FP_PROB","CONSTR_CC"] + (sum_null_found>0); 
	  gene_discov_results["TP_PROB","CONSTR_CC"] = gene_discov_results["TP_PROB","CONSTR_CC"] + (sum_nonnull_found>0);
	  gene_discov_results["REJECT_PROB","CONSTR_CC"] = gene_discov_results["REJECT_PROB","CONSTR_CC"] + (sum_all_found>0);
	  
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (3a) CONSTRAINED CC		#
	  #BEGIN (3b) CHISQ,CC						#
	  #################################################################
	  all_found = chisq_cc_stat>bonferroni_sig_chisq2;
	  sum_all_found = sum(all_found);
	  
	  #Summaries of all GxE SNPs found
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  ge_results["GV","CHISQ_CC"] = ge_results["GV","CHISQ_CC"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","CHISQ_CC"] = ge_results["FP_COUNT","CHISQ_CC"] + sum_null_found; 
	  ge_results["TP_COUNT","CHISQ_CC"] = ge_results["TP_COUNT","CHISQ_CC"] + sum_nonnull_found;
	  ge_results["FDR","CHISQ_CC"] = ge_results["FDR","CHISQ_CC"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","CHISQ_CC"] = ge_results["FNR","CHISQ_CC"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","CHISQ_CC"] = ge_results["FP_PROB","CHISQ_CC"] + (sum_null_found>0); 
	  ge_results["TP_PROB","CHISQ_CC"] = ge_results["TP_PROB","CHISQ_CC"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","CHISQ_CC"] = ge_results["REJECT_PROB","CHISQ_CC"] + (sum_all_found>0);
	  
	  sum_null_found = sum(all_found[null_gene_discov_index]);
	  sum_nonnull_found = sum(all_found[nonnull_gene_discov_index]);
	  
	  #Summaries of gene discoveries
	  gene_discov_results["GV","CHISQ_CC"] = gene_discov_results["GV","CHISQ_CC"] + sum(all_found*marg_effect_size);
	  gene_discov_results["FP_COUNT","CHISQ_CC"] = gene_discov_results["FP_COUNT","CHISQ_CC"] + sum_null_found; 
	  gene_discov_results["TP_COUNT","CHISQ_CC"] = gene_discov_results["TP_COUNT","CHISQ_CC"] + sum_nonnull_found;
	  gene_discov_results["FDR","CHISQ_CC"] = gene_discov_results["FDR","CHISQ_CC"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  gene_discov_results["FNR","CHISQ_CC"] = gene_discov_results["FNR","CHISQ_CC"] + max(0,(sum_nonnull_gene_discov-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  gene_discov_results["FP_PROB","CHISQ_CC"] = gene_discov_results["FP_PROB","CHISQ_CC"] + (sum_null_found>0); 
	  gene_discov_results["TP_PROB","CHISQ_CC"] = gene_discov_results["TP_PROB","CHISQ_CC"] + (sum_nonnull_found>0);
	  gene_discov_results["REJECT_PROB","CHISQ_CC"] = gene_discov_results["REJECT_PROB","CHISQ_CC"] + (sum_all_found>0);
	  
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (3b) CHISQ,CC						#
	  #BEGIN (3c) CHISQ,CO						#
	  #################################################################
	  all_found = chisq_co_stat>bonferroni_sig_chisq2;
	  sum_all_found = sum(all_found);
	  
	  #Summaries of all GxE SNPs found
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  ge_results["GV","CHISQ_CO"] = ge_results["GV","CHISQ_CO"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","CHISQ_CO"] = ge_results["FP_COUNT","CHISQ_CO"] + sum_null_found; 
	  ge_results["TP_COUNT","CHISQ_CO"] = ge_results["TP_COUNT","CHISQ_CO"] + sum_nonnull_found;
	  ge_results["FDR","CHISQ_CO"] = ge_results["FDR","CHISQ_CO"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","CHISQ_CO"] = ge_results["FNR","CHISQ_CO"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","CHISQ_CO"] = ge_results["FP_PROB","CHISQ_CO"] + (sum_null_found>0); 
	  ge_results["TP_PROB","CHISQ_CO"] = ge_results["TP_PROB","CHISQ_CO"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","CHISQ_CO"] = ge_results["REJECT_PROB","CHISQ_CO"] + (sum_all_found>0);
	  
	  sum_null_found = sum(all_found[null_gene_discov_index]);
	  sum_nonnull_found = sum(all_found[nonnull_gene_discov_index]);
	  
	  #Summaries of gene discoveries
	  gene_discov_results["GV","CHISQ_CO"] = gene_discov_results["GV","CHISQ_CO"] + sum(all_found*marg_effect_size);
	  gene_discov_results["FP_COUNT","CHISQ_CO"] = gene_discov_results["FP_COUNT","CHISQ_CO"] + sum_null_found; 
	  gene_discov_results["TP_COUNT","CHISQ_CO"] = gene_discov_results["TP_COUNT","CHISQ_CO"] + sum_nonnull_found;
	  gene_discov_results["FDR","CHISQ_CO"] = gene_discov_results["FDR","CHISQ_CO"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  gene_discov_results["FNR","CHISQ_CO"] = gene_discov_results["FNR","CHISQ_CO"] + max(0,(sum_nonnull_gene_discov-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  gene_discov_results["FP_PROB","CHISQ_CO"] = gene_discov_results["FP_PROB","CHISQ_CO"] + (sum_null_found>0); 
	  gene_discov_results["TP_PROB","CHISQ_CO"] = gene_discov_results["TP_PROB","CHISQ_CO"] + (sum_nonnull_found>0);
	  gene_discov_results["REJECT_PROB","CHISQ_CO"] = gene_discov_results["REJECT_PROB","CHISQ_CO"] + (sum_all_found>0);
	  
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (3c) CHISQ,CO						#
	  #BEGIN (3d) CHISQ,EB						#
	  #################################################################
	  all_found = chisq_eb_stat>bonferroni_sig_chisq2;
	  sum_all_found = sum(all_found);
	  
	  #Summaries of all GxE SNPs found
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  ge_results["GV","CHISQ_EB"] = ge_results["GV","CHISQ_EB"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","CHISQ_EB"] = ge_results["FP_COUNT","CHISQ_EB"] + sum_null_found; 
	  ge_results["TP_COUNT","CHISQ_EB"] = ge_results["TP_COUNT","CHISQ_EB"] + sum_nonnull_found;
	  ge_results["FDR","CHISQ_EB"] = ge_results["FDR","CHISQ_EB"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","CHISQ_EB"] = ge_results["FNR","CHISQ_EB"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","CHISQ_EB"] = ge_results["FP_PROB","CHISQ_EB"] + (sum_null_found>0); 
	  ge_results["TP_PROB","CHISQ_EB"] = ge_results["TP_PROB","CHISQ_EB"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","CHISQ_EB"] = ge_results["REJECT_PROB","CHISQ_EB"] + (sum_all_found>0);
	  
	  sum_null_found = sum(all_found[null_gene_discov_index]);
	  sum_nonnull_found = sum(all_found[nonnull_gene_discov_index]);
	  
	  #Summaries of gene discoveries
	  gene_discov_results["GV","CHISQ_EB"] = gene_discov_results["GV","CHISQ_EB"] + sum(all_found*marg_effect_size);
	  gene_discov_results["FP_COUNT","CHISQ_EB"] = gene_discov_results["FP_COUNT","CHISQ_EB"] + sum_null_found; 
	  gene_discov_results["TP_COUNT","CHISQ_EB"] = gene_discov_results["TP_COUNT","CHISQ_EB"] + sum_nonnull_found;
	  gene_discov_results["FDR","CHISQ_EB"] = gene_discov_results["FDR","CHISQ_EB"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  gene_discov_results["FNR","CHISQ_EB"] = gene_discov_results["FNR","CHISQ_EB"] + max(0,(sum_nonnull_gene_discov-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  gene_discov_results["FP_PROB","CHISQ_EB"] = gene_discov_results["FP_PROB","CHISQ_EB"] + (sum_null_found>0); 
	  gene_discov_results["TP_PROB","CHISQ_EB"] = gene_discov_results["TP_PROB","CHISQ_EB"] + (sum_nonnull_found>0);
	  gene_discov_results["REJECT_PROB","CHISQ_EB"] = gene_discov_results["REJECT_PROB","CHISQ_EB"] + (sum_all_found>0);
	  
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (3d) CHISQ,EB							#
	  #BEGIN (3e) KRAFT,LRT							#
	  #################################################################
	  all_found = kraft_lrt_stat>bonferroni_sig_chisq2;
	  sum_all_found = sum(all_found);
	  
	  #Summaries of all GxE SNPs found
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  ge_results["GV","KRAFT"] = ge_results["GV","KRAFT"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","KRAFT"] = ge_results["FP_COUNT","KRAFT"] + sum_null_found; 
	  ge_results["TP_COUNT","KRAFT"] = ge_results["TP_COUNT","KRAFT"] + sum_nonnull_found;
	  ge_results["FDR","KRAFT"] = ge_results["FDR","KRAFT"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","KRAFT"] = ge_results["FNR","KRAFT"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","KRAFT"] = ge_results["FP_PROB","KRAFT"] + (sum_null_found>0); 
	  ge_results["TP_PROB","KRAFT"] = ge_results["TP_PROB","KRAFT"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","KRAFT"] = ge_results["REJECT_PROB","KRAFT"] + (sum_all_found>0);
	  
	  sum_null_found = sum(all_found[null_gene_discov_index]);
	  sum_nonnull_found = sum(all_found[nonnull_gene_discov_index]);
	  
	  #Summaries of gene discoveries
	  gene_discov_results["GV","KRAFT"] = gene_discov_results["GV","KRAFT"] + sum(all_found*marg_effect_size);
	  gene_discov_results["FP_COUNT","KRAFT"] = gene_discov_results["FP_COUNT","KRAFT"] + sum_null_found; 
	  gene_discov_results["TP_COUNT","KRAFT"] = gene_discov_results["TP_COUNT","KRAFT"] + sum_nonnull_found;
	  gene_discov_results["FDR","KRAFT"] = gene_discov_results["FDR","KRAFT"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  gene_discov_results["FNR","KRAFT"] = gene_discov_results["FNR","KRAFT"] + max(0,(sum_nonnull_gene_discov-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  gene_discov_results["FP_PROB","KRAFT"] = gene_discov_results["FP_PROB","KRAFT"] + (sum_null_found>0); 
	  gene_discov_results["TP_PROB","KRAFT"] = gene_discov_results["TP_PROB","KRAFT"] + (sum_nonnull_found>0);
	  gene_discov_results["REJECT_PROB","KRAFT"] = gene_discov_results["REJECT_PROB","KRAFT"] + (sum_all_found>0);
	  
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #####################################################
	  #END (3e) KRAFT,LRT								 #
	  #BEGIN (3f) MARGINAL G TEST				#
	  #################################################################
	  all_found = screen_marg>bonferroni_sig;
	  sum_all_found = sum(all_found);
	  
	  #Summaries of all GxE SNPs found
	  sum_null_found = sum(all_found[int_type0_index]);
	  sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	  ge_results["GV","G_MARG"] = ge_results["GV","G_MARG"] + sum(all_found*marg_effect_size);
	  ge_results["FP_COUNT","G_MARG"] = ge_results["FP_COUNT","G_MARG"] + sum_null_found; 
	  ge_results["TP_COUNT","G_MARG"] = ge_results["TP_COUNT","G_MARG"] + sum_nonnull_found;
	  ge_results["FDR","G_MARG"] = ge_results["FDR","G_MARG"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  ge_results["FNR","G_MARG"] = ge_results["FNR","G_MARG"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  ge_results["FP_PROB","G_MARG"] = ge_results["FP_PROB","G_MARG"] + (sum_null_found>0); 
	  ge_results["TP_PROB","G_MARG"] = ge_results["TP_PROB","G_MARG"] + (sum_nonnull_found>0);
	  ge_results["REJECT_PROB","G_MARG"] = ge_results["REJECT_PROB","G_MARG"] + (sum_all_found>0);
	  
	  sum_null_found = sum(all_found[null_gene_discov_index]);
	  sum_nonnull_found = sum(all_found[nonnull_gene_discov_index]);
	  
	  #Summaries of gene discoveries
	  gene_discov_results["GV","G_MARG"] = gene_discov_results["GV","G_MARG"] + sum(all_found*marg_effect_size);
	  gene_discov_results["FP_COUNT","G_MARG"] = gene_discov_results["FP_COUNT","G_MARG"] + sum_null_found; 
	  gene_discov_results["TP_COUNT","G_MARG"] = gene_discov_results["TP_COUNT","G_MARG"] + sum_nonnull_found;
	  gene_discov_results["FDR","G_MARG"] = gene_discov_results["FDR","G_MARG"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	  gene_discov_results["FNR","G_MARG"] = gene_discov_results["FNR","G_MARG"] + max(0,(sum_nonnull_gene_discov-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	  gene_discov_results["FP_PROB","G_MARG"] = gene_discov_results["FP_PROB","G_MARG"] + (sum_null_found>0); 
	  gene_discov_results["TP_PROB","G_MARG"] = gene_discov_results["TP_PROB","G_MARG"] + (sum_nonnull_found>0);
	  gene_discov_results["REJECT_PROB","G_MARG"] = gene_discov_results["REJECT_PROB","G_MARG"] + (sum_all_found>0);
	  
	  rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	  
	  #################################################################
	  #END (3f) MARGINAL G TEST				#
	  if(include_constrained) {
	    #BEGIN (3g) CONSTRAINED CC (CML)   		#
	    #################################################################	    
	    all_found = cnstr_cc_cml_stat>bonferroni_sig_chisq1
	    sum_all_found = sum(all_found);
	    
	    #Summaries of all GxE SNPs found
	    sum_null_found = sum(all_found[int_type0_index]);
	    sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	    ge_results["GV","CONSTR_CC_CML"] = ge_results["GV","CONSTR_CC_CML"] + sum(all_found*marg_effect_size);
	    ge_results["FP_COUNT","CONSTR_CC_CML"] = ge_results["FP_COUNT","CONSTR_CC_CML"] + sum_null_found; 
	    ge_results["TP_COUNT","CONSTR_CC_CML"] = ge_results["TP_COUNT","CONSTR_CC_CML"] + sum_nonnull_found;
	    ge_results["FDR","CONSTR_CC_CML"] = ge_results["FDR","CONSTR_CC_CML"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	    ge_results["FNR","CONSTR_CC_CML"] = ge_results["FNR","CONSTR_CC_CML"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	    ge_results["FP_PROB","CONSTR_CC_CML"] = ge_results["FP_PROB","CONSTR_CC_CML"] + (sum_null_found>0); 
	    ge_results["TP_PROB","CONSTR_CC_CML"] = ge_results["TP_PROB","CONSTR_CC_CML"] + (sum_nonnull_found>0);
	    ge_results["REJECT_PROB","CONSTR_CC_CML"] = ge_results["REJECT_PROB","CONSTR_CC_CML"] + (sum_all_found>0);
	    
	    sum_null_found = sum(all_found[null_gene_discov_index]);
	    sum_nonnull_found = sum(all_found[nonnull_gene_discov_index]);
	    
	    #Summaries of gene discoveries
	    gene_discov_results["GV","CONSTR_CC_CML"] = gene_discov_results["GV","CONSTR_CC_CML"] + sum(all_found*marg_effect_size);
	    gene_discov_results["FP_COUNT","CONSTR_CC_CML"] = gene_discov_results["FP_COUNT","CONSTR_CC_CML"] + sum_null_found; 
	    gene_discov_results["TP_COUNT","CONSTR_CC_CML"] = gene_discov_results["TP_COUNT","CONSTR_CC_CML"] + sum_nonnull_found;
	    gene_discov_results["FDR","CONSTR_CC_CML"] = gene_discov_results["FDR","CONSTR_CC_CML"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	    gene_discov_results["FNR","CONSTR_CC_CML"] = gene_discov_results["FNR","CONSTR_CC_CML"] + max(0,(sum_nonnull_gene_discov-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	    gene_discov_results["FP_PROB","CONSTR_CC_CML"] = gene_discov_results["FP_PROB","CONSTR_CC_CML"] + (sum_null_found>0); 
	    gene_discov_results["TP_PROB","CONSTR_CC_CML"] = gene_discov_results["TP_PROB","CONSTR_CC_CML"] + (sum_nonnull_found>0);
	    gene_discov_results["REJECT_PROB","CONSTR_CC_CML"] = gene_discov_results["REJECT_PROB","CONSTR_CC_CML"] + (sum_all_found>0);
	    
	    rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	    
	    #################################################################
	    #END (3g) CONSTRAINED CC (CML) 				#
	    #BEGIN (3h) KRAFT (CML)			#
	    #################################################################	 
	    all_found = kraft_cml_stat>bonferroni_sig_chisq2;
	    sum_all_found = sum(all_found);
	    
	    #Summaries of all GxE SNPs found
	    sum_null_found = sum(all_found[int_type0_index]);
	    sum_nonnull_found = sum(all_found[nonnull_OR_GxE_index]);
	    ge_results["GV","KRAFT_CML"] = ge_results["GV","KRAFT_CML"] + sum(all_found*marg_effect_size);
	    ge_results["FP_COUNT","KRAFT_CML"] = ge_results["FP_COUNT","KRAFT_CML"] + sum_null_found; 
	    ge_results["TP_COUNT","KRAFT_CML"] = ge_results["TP_COUNT","KRAFT_CML"] + sum_nonnull_found;
	    ge_results["FDR","KRAFT_CML"] = ge_results["FDR","KRAFT_CML"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	    ge_results["FNR","KRAFT_CML"] = ge_results["FNR","KRAFT_CML"] + max(0,(M-sum_null_OR_GxE-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	    ge_results["FP_PROB","KRAFT_CML"] = ge_results["FP_PROB","KRAFT_CML"] + (sum_null_found>0); 
	    ge_results["TP_PROB","KRAFT_CML"] = ge_results["TP_PROB","KRAFT_CML"] + (sum_nonnull_found>0);
	    ge_results["REJECT_PROB","KRAFT_CML"] = ge_results["REJECT_PROB","KRAFT_CML"] + (sum_all_found>0);
	    
	    sum_null_found = sum(all_found[null_gene_discov_index]);
	    sum_nonnull_found = sum(all_found[nonnull_gene_discov_index]);
	    
	    #Summaries of gene discoveries
	    gene_discov_results["GV","KRAFT_CML"] = gene_discov_results["GV","KRAFT_CML"] + sum(all_found*marg_effect_size);
	    gene_discov_results["FP_COUNT","KRAFT_CML"] = gene_discov_results["FP_COUNT","KRAFT_CML"] + sum_null_found; 
	    gene_discov_results["TP_COUNT","KRAFT_CML"] = gene_discov_results["TP_COUNT","KRAFT_CML"] + sum_nonnull_found;
	    gene_discov_results["FDR","KRAFT_CML"] = gene_discov_results["FDR","KRAFT_CML"] + max(0,sum_null_found/sum_all_found,na.rm=T);
	    gene_discov_results["FNR","KRAFT_CML"] = gene_discov_results["FNR","KRAFT_CML"] + max(0,(sum_nonnull_gene_discov-sum_nonnull_found)/(M-sum_all_found),na.rm=T);
	    gene_discov_results["FP_PROB","KRAFT_CML"] = gene_discov_results["FP_PROB","KRAFT_CML"] + (sum_null_found>0); 
	    gene_discov_results["TP_PROB","KRAFT_CML"] = gene_discov_results["TP_PROB","KRAFT_CML"] + (sum_nonnull_found>0);
	    gene_discov_results["REJECT_PROB","KRAFT_CML"] = gene_discov_results["REJECT_PROB","KRAFT_CML"] + (sum_all_found>0);
	    
	    rm(all_found,sum_all_found,sum_null_found,sum_nonnull_found);
	    #################################################################
	    #END (3h) KRAFT (CML)			#
	    #################################################################	 
    }
    
    
	  if(i%%report_freq==0) {cat(paste("(",sim_number,"-",i,")",sep=""),"-- proj. run time:",round(proj.run.time<-niter*difftime(Sys.time(),begin,units=report_units)/i,3),report_units,"-- proj. end time: ",format.Date(begin  + proj.run.time),"\n\n",file="results.txt",append=T)}
	  
	}
  
  #Parameters used that are common to both random and fixed parameter settings.
  common_params = c(sim_number,niter,ncase,ncontrol,nsnp, 
                    gw_alpha,screen_alpha_lax,screen_alpha_rigid,
                    sens_E_case,spec_E_case,
                    sens_E_control,spec_E_control,
                    prob_E,marg_OR_E[1]);
  
  names(common_params) = c("sim_number","niter","ncase","ncontrol","nsnp", 
                           "gw_alpha","screen_alpha_lax","screen_alpha_rigid",
                           "sens_E_case","spec_E_case",
                           "sens_E_control","spec_E_control",
                           "prob_E","marg_OR_E");
  
  #Parameters used that are specific to fixed parameter setting
  if(!random_ORs) {causal_OR_E = OR_E[1];causal_marg_OR_G = marg_OR_G[1];}
  fixed_params = c(causal_maf, prob_GxE_ind, GxE_assc_sd,
                   causal_OR_GxE_assc, null_OR_E, causal_OR_E, 
                   num_null_with_nonnull_OR_G, causal_OR_G,causal_OR_GxE,causal_marg_OR_G);
  
  names(fixed_params) = c("causal_maf","prob_GxE_ind","GxE_assc_sd",
                          "causal_OR_GxE_assc","null_OR_E","causal_OR_E",
                          "num_null_with_nonnull_OR_G","causal_OR_G","causal_OR_GxE","causal_marg_OR_G");
  
  ge_results = ge_results/niter;
  gene_discov_results = gene_discov_results/niter;
  #Storey's Positive FDR: FDR given that at least one discovery is made
  ge_results["POS_FDR",] = ge_results["FDR",]/ge_results["REJECT_PROB",];
  
  sum_effect_size = sum_effect_size/niter;
  
  nsnps_passed_screening = nsnps_passed_screening/niter;
  ge_snp_types = ge_snp_types/niter;  
  gene_discov_snp_types = gene_discov_snp_types/niter; 
  
  return(list(common_params = common_params,
              fixed_params = fixed_params,
              OR_GxE_assc_params = OR_GxE_assc_params,
              marg_OR_G_params = marg_OR_G_params,
              ge_snp_types = ge_snp_types,
              gene_discov_snp_types = gene_discov_snp_types, 
              sum_effect_size = sum_effect_size,
              ge_results=ge_results, 
              gene_discov_results=gene_discov_results, 
              nsnps_passed_screening = nsnps_passed_screening,
              logic_check=logic_check,
              GE_DG_cor_null=GE_DG_cor_null,
              GE_DG_cor_all=GE_DG_cor_all));
}

